/*******************************************************************************
    Copyright (c) 2013 NVidia Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_ioctl.h"
#include "uvm_linux_ioctl.h"
#include "nvidia_uvm_common.h"
#include "uvm_gpu_provider.h"
#include "nvidia_page_migration.h"
#include "uvm_gpu_provider.h"
#include "nvidia_uvm_utils.h"
#include "nvidia_uvm_lite.h"
#include "nvidia_uvm_lite_counters.h"

//
// nvidia_uvm_lite.c
// This file contains code that is specific to the UVM-Lite mode of operation.
//

#ifndef NVIDIA_UVM_LITE_ENABLED
#error "Building uvmlite code without NVIDIA_UVM_LITE_ENABLED!"
#endif
#ifndef NVIDIA_UVM_RM_ENABLED
#error "Building uvmlite code without RM enabled!"
#endif


static struct cdev g_uvmlite_cdev;

// table of attached GUIDS
static UvmGpuUuid g_attached_uuid_list[UVM_MAX_GPUS];

//
// Locking acquisition order (only take the locks you need, but always follow
// this order):
//
//      1. mm->mmap_sem
//
//      2. DriverPrivate.uvmPrivLock
//
//      3. UvmProcessCounterInfo.gpuHistoryLock
//
//      4. g_uvmProcessRecordTableLock
//
//      5. sessionInfoLock
//
static struct kmem_cache * g_uvmPrivateCache              __read_mostly = NULL;
static struct kmem_cache * g_uvmCommitRecordCache         __read_mostly = NULL;
static struct kmem_cache * g_uvmMigTrackerCache           __read_mostly = NULL;
static struct kmem_cache * g_uvmProcessCounterInfoCache   __read_mostly = NULL;
static struct kmem_cache * g_uvmStreamRecordCache         __read_mostly = NULL;
static struct kmem_cache * g_uvmMappingCache              __read_mostly = NULL;

static struct rw_semaphore g_uvmProcessRecordTableLock;
//
// Root of global process record list
// This list contains UvmProcessRecord pointers,
// which are valid as long, as you are holding
// r/w g_uvmProcessRecordTableLock
//
static LIST_HEAD(g_uvmProcessRecordList);


static void _mmap_close(struct vm_area_struct *vma);
static void _mmap_open(struct vm_area_struct *vma);
static void _destroy_migration_resources(UvmGpuMigrationTracking *pMigTracking);
static int _create_migration_resources(UvmGpuUuid * pGpuUuidStruct,
                                       UvmGpuMigrationTracking *pMigTracking);
static void _set_vma_inaccessible(struct vm_area_struct * vma);
static void _set_vma_accessible(struct vm_area_struct * vma);
static void _destroy_record(UvmCommitRecord *pRecord);
static void _set_record_accessible(UvmCommitRecord *pRecord);
static void _set_record_inaccessible(UvmCommitRecord *pRecord);
static int _is_record_matching_vma(UvmCommitRecord *pRecord);
static void _record_detach_from_stream(UvmCommitRecord *pRecord);
static UVM_STATUS _record_attach_to_stream(UvmCommitRecord *pRecord,
    UvmStreamRecord *pNewStream);

static UvmStreamRecord * _stream_alloc(UvmProcessRecord *processRecord,
    UvmStream streamId);
static UvmStreamRecord * _stream_find_in_cache(UvmProcessRecord *processRecord,
    UvmStream streamId);
static UvmStreamRecord * _stream_find_in_list(UvmProcessRecord *processRecord,
    UvmStream streamId);
static UvmStreamRecord * _stream_find(UvmProcessRecord *processRecord,
    UvmStream streamId);
static UvmStreamRecord * _stream_find_or_alloc(UvmProcessRecord *processRecord,
    UvmStream streamId);
static void _stream_destroy(UvmStreamRecord *pStream);
static void _stream_destroy_if_empty(UvmStreamRecord *pStream);
static void _stream_remove_from_cache(UvmStreamRecord *pStream);
static void _stream_save_in_cache(UvmStreamRecord *pStream);

//
// Currently the driver refuses to work with vmas that have been modified
// since the original mmap() call. Mark them as inaccessible.
//
// Locking: this is called with mmap_sem for write
//
static void _mmap_open(struct vm_area_struct *vma)
{
    UvmCommitRecord *pRecord = (UvmCommitRecord*)vma->vm_private_data;

    pr_debug_ratelimited("mmap_open(pid:%u): vma 0x%p [0x%p, 0x%p) record %p\n",
        current->pid,
        vma, (void*)vma->vm_start, (void*)vma->vm_end,
        pRecord);

    //
    // The vma that was originally created is being modified.
    // Mark the cloned vma as inaccessible and reset its private data to make
    // sure the same commit record is not referenced by multiple vmas.
    //
    _set_vma_inaccessible(vma);
    vma->vm_private_data = NULL;

    if (pRecord)
    {
        DriverPrivate *pDriverPriv = pRecord->osPrivate;

        // mmap_open should never be called for the original vma
        WARN_ON(pRecord->vma == vma);

        // Set the original vma as inaccessible
        _set_vma_inaccessible(pRecord->vma);

        // Destroy the commit record attached to the original vma
        down_write(&pDriverPriv->uvmPrivLock);
        _destroy_record(pRecord);
        up_write(&pDriverPriv->uvmPrivLock);
    }
}

static void _stream_destroy(UvmStreamRecord *pStream)
{
    if (pStream == NULL)
        return;

    pr_debug_ratelimited("%s(pid:%u): stream %lld\n",
                         __FUNCTION__, current->pid, pStream->streamId);

    // Stream should be stopped
    WARN_ON(pStream->isRunning);

    // Stream should be empty
    WARN_ON(!list_empty(&pStream->commitRecordsList));

    list_del(&pStream->allStreamListNode);
    _stream_remove_from_cache(pStream);

    kmem_cache_free(g_uvmStreamRecordCache, pStream);
}

void _stop_and_destroy_leftover_streams(UvmProcessRecord *processRecord)
{
    struct list_head *pos;
    struct list_head *safepos;
    UvmStreamRecord *pStream = NULL;

    list_for_each_safe(pos, safepos, &processRecord->allStreamList)
    {
        pStream = list_entry(pos, UvmStreamRecord, allStreamListNode);
        pStream->isRunning = UVM_FALSE;
        _stream_destroy(pStream);
    }
}

//
// Locking: you must hold a write lock on the DriverPrivate.uvmPrivLock
// and mmap_sem, before calling this routine.
//
static void _destroy_record(UvmCommitRecord *pRecord)
{
    if (NULL == pRecord)
        return;

    vfree(pRecord->commitRecordPages);
    pRecord->commitRecordPages = NULL;

    _record_detach_from_stream(pRecord);

    if (pRecord->vma)
    {
        // Detach the record from the vma
        pRecord->vma->vm_private_data = NULL;
        pRecord->vma = NULL;
    }

    kmem_cache_free(g_uvmCommitRecordCache, pRecord);
}

static void _mmap_close(struct vm_area_struct *vma)
{
    UvmCommitRecord *pRecord = (UvmCommitRecord*)vma->vm_private_data;

    pr_debug_ratelimited(
        "mmap_close(pid:%u): vma 0x%p [0x%p, 0x%p) record 0x%p\n",
        current->pid,
        vma, (void*)vma->vm_start, (void*)vma->vm_end,
        pRecord);

    if (pRecord)
    {
        DriverPrivate *pDriverPriv = pRecord->osPrivate;

        //
        // This should never happen as the vm_private_data is reset
        // in mmap_open().
        //
        WARN_ON(pRecord->vma != vma);

        down_write(&pDriverPriv->uvmPrivLock);
        _destroy_record(pRecord);
        up_write(&pDriverPriv->uvmPrivLock);
    }
}

//
// Locking: you must hold a write lock on the DriverPrivate.uvmPrivLock, before
// calling this routine.
//
// returns a free index or an index of a matching mig record on success
// returns negative on failure
//
static int _find_free_or_matching_per_process_mig_record(
    const UvmCommitRecord * pRecord)
{
    int index;
    UvmPerProcessGpuMigs * pMig;

    WARN_ON((pRecord  == NULL) || (pRecord->osPrivate == NULL));

    if ((pRecord  == NULL) || (pRecord->osPrivate == NULL))
        return -EINVAL;

    for (index = 0; index < UVM_MAX_GPUS; ++index)
    {
        // The gpuMigs array will never have any holes as
        // we always add to its end and never remove anything
        // except when clearing it completely.
        pMig = &pRecord->osPrivate->processRecord.gpuMigs[index];

        if (pMig->migTracker == NULL)
        {
            // Finding a NULL tracker indicates the end of the occupied slots.
            // Claim and return the free slot.
            memcpy(&(pMig->migUuid.uuid), pRecord->homeGpuUuid.uuid,
                   sizeof(pMig->migUuid.uuid));
            return index;
        }

        if (memcmp(&(pMig->migUuid.uuid), pRecord->homeGpuUuid.uuid,
                sizeof(pMig->migUuid.uuid)) == 0)
        {
            // Found a matching tracker.
            return index;
        }
    }

    return -ENOENT;
}

//
// Locking: you must hold a write lock on the DriverPrivate.uvmPrivLock, before
// calling this routine.
//
static void _disconnect_mig_completely(UvmPerProcessGpuMigs * pMig,
                                       UvmCommitRecord * pRecord)
{
    memset(pMig, 0, sizeof(*pMig));

    if (pRecord != NULL)
    {
        pRecord->cachedHomeGpuPerProcessIndex = UVM_INVALID_HOME_GPU_INDEX;
        pRecord->cachedPerProcessPerGpuCounterIndex =
                                  UVM_INVALID_PER_PROCESS_PER_GPU_COUNTER_INDEX;
    }
}

static void
_delete_all_session_info_table_entries(UvmProcessRecord *pProcessRecord)
{
    int i;

    down_write(&pProcessRecord->sessionInfoLock);

    for (i = 0; i < UVM_MAX_SESSIONS_PER_PROCESS; ++i)
    {
        if (pProcessRecord->sessionInfoArray[i].pidSessionOwner != 
            UVM_PID_INIT_VALUE)
        {
            UvmSessionInfo *pSessionInfo = &pProcessRecord->sessionInfoArray[i];
            //
            // Free counter information if this is the only process using it.
            // This would happen if the target process has exited.
            //
            uvmlite_put_and_unrefcount_process_counter_info(
                pSessionInfo->pTargetCounterInfo);

            UvmInitSessionInfo(pSessionInfo);
        }
    }

    up_write(&pProcessRecord->sessionInfoLock);
}

//
// Locking: you must hold a write lock on the DriverPrivate.uvmPrivLock, before
// calling this routine.
//
static void
_delete_all_migration_resources(DriverPrivate* pPriv)
{
    int index;
    UvmPerProcessGpuMigs * pMig;
    char uuidBuffer[UVM_GPU_UUID_TEXT_BUFFER_LENGTH];

    for (index = 0; index < UVM_MAX_GPUS; ++index)
    {
        pMig = &pPriv->processRecord.gpuMigs[index];
        if (pMig->migTracker != NULL)
        {
            format_uuid_to_buffer(uuidBuffer, sizeof(uuidBuffer),
                                  &pMig->migUuid);

            pr_debug_ratelimited(
                "%s(pid:%u): %s: (channelClass: 0x%0x, ceClass: 0x%0x)\n",
                __FUNCTION__, current->pid, uuidBuffer,
                pMig->migTracker->channelInfo.channelClassNum,
                pMig->migTracker->ceClassNumber);

            _destroy_migration_resources(pMig->migTracker);
            kmem_cache_free(g_uvmMigTrackerCache, pMig->migTracker);
            _disconnect_mig_completely(pMig, NULL);
        }
    }
}

//
// Locking: you must hold a write lock on the DriverPrivate.uvmPrivLock, before
// calling this routine.
//
static int _create_or_check_channel(UvmCommitRecord * pRecord)
{
    UvmPerProcessGpuMigs *pMig;
    int index, result;

    index = _find_free_or_matching_per_process_mig_record(pRecord);
    if (index < 0)
    {
        result = index;
        goto fail2;
    }

    pRecord->cachedHomeGpuPerProcessIndex = index;

    pMig = &pRecord->osPrivate->processRecord.gpuMigs[index];
    if (pMig->migTracker != NULL)
        // Reusing an already created mig tracker
        return 0;

    // Got a free slot, need to create the first mig tracker for this gpu
    pMig->migTracker = kmem_cache_alloc(g_uvmMigTrackerCache, GFP_KERNEL);
    if (pMig->migTracker == NULL)
    {
        result = -ENOMEM;
        goto fail2;
    }

    memset(pMig->migTracker, 0, sizeof(*pMig->migTracker));

    result = _create_migration_resources(&pRecord->homeGpuUuid,
                                         pMig->migTracker);
    if (0 != result)
        goto fail;

    return result;

fail:
    kmem_cache_free(g_uvmMigTrackerCache, pMig->migTracker);
    _disconnect_mig_completely(pMig, pRecord);
fail2:
    pRecord->cachedHomeGpuPerProcessIndex = UVM_INVALID_HOME_GPU_INDEX;
    pRecord->cachedPerProcessPerGpuCounterIndex =
                                UVM_INVALID_PER_PROCESS_PER_GPU_COUNTER_INDEX;
    return result;
}

static UvmStreamRecord * _stream_find_in_cache(UvmProcessRecord *processRecord,
    UvmStream streamId)
{
    UvmStreamRecord *stream;
    int cacheIndex = streamId % UVM_STREAMS_CACHE_SIZE;

    stream = processRecord->streamsCache[cacheIndex];

    if (stream && stream->streamId == streamId)
        return stream;

    return NULL;
}

static void _stream_remove_from_cache(UvmStreamRecord *pStream)
{
    int cacheIndex = pStream->streamId % UVM_STREAMS_CACHE_SIZE;
    if (pStream->processRecord->streamsCache[cacheIndex] == pStream)
        pStream->processRecord->streamsCache[cacheIndex] = NULL;
}

static void _stream_save_in_cache(UvmStreamRecord *pStream)
{
    int cacheIndex;

    if (!pStream)
        return;

    cacheIndex = pStream->streamId % UVM_STREAMS_CACHE_SIZE;
    pStream->processRecord->streamsCache[cacheIndex] = pStream;
}

static UvmStreamRecord * _stream_find_in_list(UvmProcessRecord *processRecord,
    UvmStream streamId)
{
    struct list_head *pos;
    UvmStreamRecord *pStream = NULL;

    list_for_each(pos, &processRecord->allStreamList)
    {
        pStream = list_entry(pos, UvmStreamRecord, allStreamListNode);
        if (pStream->streamId == streamId)
            return pStream;
    }

    return NULL;
}

static UvmStreamRecord * _stream_alloc(UvmProcessRecord *processRecord,
    UvmStream streamId)
{
    UvmStreamRecord *pStream = NULL;

    pr_debug_ratelimited("%s(pid:%u): stream %lld\n",
                         __FUNCTION__, current->pid, streamId);

    pStream = (UvmStreamRecord*)kmem_cache_alloc(g_uvmStreamRecordCache,
                                                 GFP_KERNEL);
    if (pStream == NULL)
        return NULL;

    memset(pStream, 0, sizeof(*pStream));
    INIT_LIST_HEAD(&pStream->allStreamListNode);
    INIT_LIST_HEAD(&pStream->commitRecordsList);

    pStream->processRecord = processRecord;
    pStream->streamId       = streamId;
    pStream->isRunning      = UVM_FALSE;

    list_add_tail(&pStream->allStreamListNode, &processRecord->allStreamList);

    return pStream;
}

static UvmStreamRecord * _stream_find(UvmProcessRecord *processRecord,
    UvmStream streamId)
{
    UvmStreamRecord *pStream = _stream_find_in_cache(processRecord, streamId);

    if (pStream == NULL)
        pStream = _stream_find_in_list(processRecord, streamId);

    return pStream;
}

static UvmStreamRecord * _stream_find_or_alloc(UvmProcessRecord *processRecord,
    UvmStream streamId)
{
    UvmStreamRecord *pStream = _stream_find(processRecord, streamId);

    if (pStream == NULL)
        pStream = _stream_alloc(processRecord, streamId);

    if (pStream == NULL)
        return NULL;

    _stream_save_in_cache(pStream);

    return pStream;
}

static void _record_detach_from_stream(UvmCommitRecord *pRecord)
{
    list_del(&pRecord->streamRegionsListNode);
    _stream_destroy_if_empty(pRecord->pStream);
    pRecord->pStream = NULL;
}

static UVM_STATUS _record_attach_to_stream(UvmCommitRecord *pRecord,
    UvmStreamRecord *pNewStream)
{
    UVM_STATUS status = UVM_OK;
    UvmStreamRecord *pOldStream = pRecord->pStream;
    UvmBool runningStateChanged = UVM_TRUE;

    if (pOldStream && (pOldStream->isRunning == pNewStream->isRunning))
    {
        // No need to change the state if the record's old stream is in
        // the same state as the new stream.
        runningStateChanged = UVM_FALSE;
    }

    if (runningStateChanged)
    {
        if (pNewStream->isRunning)
        {
            _set_record_inaccessible(pRecord);
            //
            // Attaching to a running stream from a stopped stream needs
            // to trigger migration to the gpu.
            //
            status = uvmlite_migrate_to_gpu(pRecord->baseAddress,
                                               pRecord->length,
                                               0, // flags
                                               pRecord->vma,
                                               pRecord);
        }
        else
            _set_record_accessible(pRecord);
    }

    list_del(&pRecord->streamRegionsListNode);
    list_add_tail(&pRecord->streamRegionsListNode,
        &pNewStream->commitRecordsList);
    pRecord->pStream = pNewStream;

    _stream_destroy_if_empty(pOldStream);

    return status;
}

//
// Locking: you must hold a write lock on the DriverPrivate.uvmPrivLock
// and mmap_sem, before calling this routine.
//
UvmCommitRecord *
uvmlite_create_commit_record(UVM_REGION_COMMIT_PARAMS *pParams,
                             DriverPrivate *pPriv,
                             struct vm_area_struct *vma)

{
    UvmCommitRecord *pRecord = NULL;
    UvmStreamRecord *pStream = NULL;
    int result;
    // The commitRecordPages array stores one pointer to each page:
    UvmSize arrayByteLen = sizeof(pRecord->commitRecordPages[0])
                                    * (PAGE_ALIGN(pParams->length) / PAGE_SIZE);

    pRecord = (UvmCommitRecord*)kmem_cache_alloc(g_uvmCommitRecordCache,
                                                 GFP_KERNEL);
    if (NULL == pRecord)
        goto fail;

    memset(pRecord, 0, sizeof(UvmCommitRecord));
    // Be sure to initialize the list, so that _destroy_record always works:
    INIT_LIST_HEAD(&pRecord->streamRegionsListNode);
    pRecord->baseAddress = pParams->requestedBase;
    pRecord->length      = pParams->length;
    pRecord->osPrivate   = pPriv;
    pRecord->vma         = vma;
    pRecord->cachedHomeGpuPerProcessIndex = UVM_INVALID_HOME_GPU_INDEX;
    pRecord->cachedPerProcessPerGpuCounterIndex =
                            UVM_INVALID_PER_PROCESS_PER_GPU_COUNTER_INDEX;

    pRecord->commitRecordPages = vmalloc(arrayByteLen);
    if (!pRecord->commitRecordPages)
    {
        pr_err("%s(pid:%u): vmalloc(%llu) failed.\n", __FUNCTION__,
               current->pid, (unsigned long long)arrayByteLen);
        goto fail;
    }

    memset(pRecord->commitRecordPages, 0, arrayByteLen);

    memcpy(pRecord->homeGpuUuid.uuid, pParams->gpuUuid.uuid,
           sizeof(pRecord->homeGpuUuid.uuid));
    //
    // The resulting resources from this call are cleaned up at process exit, as
    // part of the _destroy_migration_resources() call:
    //
    result = _create_or_check_channel(pRecord);
    if (0 != result)
    {
        pr_err("%s(pid:%u): _create_or_check_channel "
                             "failed.\n", __FUNCTION__, current->pid);
        goto fail;
    }

    if (UVM_OK != UvmGetGpuIndex(pPriv->processRecord.pProcessCounterInfo,
              &pParams->gpuUuid, &pRecord->cachedPerProcessPerGpuCounterIndex))
        goto fail;


    pr_debug_ratelimited("%s(pid:%u): pPriv: 0x%p created pRecord: 0x%p,"
                         "arrayLen: %llu\n",
                         __FUNCTION__, current->pid, pPriv, pRecord,
                         arrayByteLen);

    pStream = _stream_find_or_alloc(&pPriv->processRecord, pParams->streamId);
    if (!pStream)
        goto fail;

    _record_attach_to_stream(pRecord, pStream);

    return pRecord;

fail:
    _destroy_record(pRecord);
    return NULL;
}

//
// Locking: you must hold a write lock on the DriverPrivate.uvmPrivLock, before
// calling this routine.
//
// CAUTION: returns NULL upon failure. This failure can occur if:
//
//     a) there is no valid migration tracking resource available.
//
//     b) The UvmCommitRecord argument is corrupt (NULL osPrivate pointer).
//
static UvmGpuMigrationTracking * _get_mig_tracker(UvmCommitRecord * pRecord)
{
    unsigned migIndex;

    WARN_ON((pRecord == NULL) || (pRecord->osPrivate == NULL));

    if ((pRecord == NULL) || (pRecord->osPrivate == NULL))
        return NULL;

    migIndex = pRecord->cachedHomeGpuPerProcessIndex;
    if (migIndex == UVM_INVALID_HOME_GPU_INDEX)
        return NULL;

    return pRecord->osPrivate->processRecord.gpuMigs[migIndex].migTracker;
}

//
// Page fault handler for UVM-Lite
//
// Locking:
//     1. the kernel calls into this routine with the mmap_sem lock taken.
//     2. this routine acquires a write lock on the DriverPrivate.uvmPrivLock.
//
// Write fault algorithm:
//
//      1. Lookup address in page tracking
//      2. If page exists, mark as dirty, and we are done.
//      3. else, do "read fault" (below), then mark the page as dirty.
//
// Read fault algorithm:
//
//      1. Lookup which stream owns the address
//      2. See if that stream is active: if so, UVM-Lite rules are violated, so
//         return SIGBUS.
//      3. Otherwise, map in a page from the cache, and allow access.
//
int _fault_common(struct vm_area_struct *vma, unsigned long vaddr,
                  struct page **ppage, unsigned vmfFlags)
{
    int retValue = VM_FAULT_SIGBUS;
    UvmCommitRecord *pRecord;
    DriverPrivate *pPriv;
    UVM_STATUS uvmStatus;
    UvmGpuMigrationTracking *pMigTracker;
    UvmUPtr cpuPhysAddr;
    unsigned long pageIndex = (vaddr - vma->vm_start) >> PAGE_SHIFT;
    UvmCounterInfo *procPerGpuCounter;
    UvmPageTracking *pTracking = NULL;

    pRecord = (UvmCommitRecord*)vma->vm_private_data;
    if (!pRecord)
        goto done;

    pPriv = pRecord->osPrivate;
    WARN_ON(!pPriv);
    if (!pPriv)
        goto done;

    down_write(&pPriv->uvmPrivLock);

    if (_is_record_matching_vma(pRecord) == UVM_FALSE)
    {
        //
        // The VMA has been modified since the record was created, skip it.
        // This should never be possible as we destroy the records with
        // modified vmas in mmap_open().
        //
        WARN_ON(1);
        goto fail;
    }

    pTracking = pRecord->commitRecordPages[pageIndex];

    pr_debug_ratelimited("FAULT_ENTRY(pid:%u): vaddr: 0x%p, "
                         "pPriv: 0x%p, vma: 0x%p, pRecord: 0x%p\n",
                         current->pid, (void*)vaddr, pPriv, vma, pRecord);
    //
    // TODO: handle write faults (this only does read faults properly).
    //       (vma->vm_flags & VM_WRITE). At that point, remove this BUG_ON:
    //
    BUG_ON(pTracking);
    if (!pTracking)
    {
        pTracking = uvm_page_cache_alloc_page(pPriv);
        if (!pTracking)
        {
            retValue = VM_FAULT_OOM;
            goto fail;
        }

        pr_debug_ratelimited("FAULT_ALLOC(pid:%u): vaddr: 0x%p, page: 0x%p, "
                             "pPriv: 0x%p, vma: 0x%p, pRecord: 0x%p\n",
                             current->pid, (void*)vaddr, pTracking->uvmPage,
                             pPriv, vma, pRecord);

        pRecord->commitRecordPages[pageIndex] = pTracking;
    }

    cpuPhysAddr = page_to_phys(pTracking->uvmPage);

    pMigTracker = _get_mig_tracker(pRecord);
    WARN_ON(!pMigTracker);
    if (!pMigTracker)
        goto fail;

    uvmStatus = migrate_gpu_to_cpu(pMigTracker, (UvmGpuPointer) vaddr,
                                   cpuPhysAddr,PAGE_SIZE);
    if (uvmStatus != UVM_OK)
    {
        pr_err("FAULT(pid:%u): failed to copy from cpu to gpu:"
               " vaddr:0x%p, page: 0x%p, physAddr:0x%p, vma: 0x%p, "
               "uvmStatus: 0x%0x\n", current->pid,
               (void*)vaddr, pTracking->uvmPage, (void*)cpuPhysAddr,
               vma, uvmStatus);
        goto fail;
    }

    pr_debug_ratelimited("FAULT HANDLED(pid:%u): vaddr: 0x%p, vma: 0x%p, "
                         "page: 0x%p (pfn:0x%lx, refcount: %d)\n",
                         current->pid, (void*)vaddr, vma, pTracking->uvmPage,
                         page_to_pfn(pTracking->uvmPage),
                         page_count(pTracking->uvmPage));

    WARN_ON(pRecord->cachedPerProcessPerGpuCounterIndex ==
                     UVM_INVALID_PER_PROCESS_PER_GPU_COUNTER_INDEX);

    procPerGpuCounter =
        &pPriv->processRecord.pProcessCounterInfo->procSingleGpuCounterArray
                  [pRecord->cachedPerProcessPerGpuCounterIndex];

    UvmIncrementProcessCounters(procPerGpuCounter, &pPriv->processRecord,
        UvmCounterNameBytesXferDtH, PAGE_SIZE);
    UvmIncrementProcessCounters(procPerGpuCounter, &pPriv->processRecord,
        UvmCounterNameCpuPageFaultCount, 1);

    lock_page(pTracking->uvmPage);
    *ppage = pTracking->uvmPage;
    retValue = VM_FAULT_LOCKED;
    up_write(&pPriv->uvmPrivLock);
    return retValue;

fail:
    if (pTracking)
        uvm_page_cache_free_page(pTracking, pPriv, __FUNCTION__);

    up_write(&pPriv->uvmPrivLock);
done:
    return retValue;
}

#if defined(NV_VM_OPERATIONS_STRUCT_HAS_FAULT)
int _fault(struct vm_area_struct *vma, struct vm_fault *vmf)
{
    unsigned long vaddr = (unsigned long)vmf->virtual_address;
    struct page *page = NULL;
    int retval;

    retval = _fault_common(vma, vaddr, &page, vmf->flags);

    vmf->page = page;

    return retval;
}

#else
struct page * _fault_old_style(struct vm_area_struct *vma,
                               unsigned long address, int *type)
{
    unsigned long vaddr = address;
    struct page *page = NULL;

    _fault_common(vma, vaddr, &page, FAULT_FLAG_FROM_OLD_KERNEL);
    return page;
}
#endif

static struct vm_operations_struct uvmlite_vma_ops =
{
    .open = _mmap_open,
    .close = _mmap_close,

#if defined(NV_VM_OPERATIONS_STRUCT_HAS_FAULT)
    .fault = _fault,
#else
    .nopage = _fault_old_style,
#endif
};

//
// Locking: you must hold a write lock on the DriverPrivate.uvmPrivLock, before
// calling this routine.
//
UVM_STATUS uvmlite_migrate_to_gpu(unsigned long long baseAddress,
                                  UvmSize length,
                                  unsigned migrateFlags,
                                  struct vm_area_struct *vma,
                                  UvmCommitRecord * pRecord)
{
    UVM_STATUS uvmStatus = UVM_OK;
    UvmPageTracking * pTracking = NULL;
    UvmUPtr cpuPhysAddr;
    UvmGpuMigrationTracking *pMigTracker;
    UvmUPtr end;
    UvmUPtr pageVirtualAddr;
    unsigned long pageIndex = 0;
    unsigned long long migratedPages = 0;
    UvmCounterInfo *procPerGpuCounter;

    if (!pRecord || !vma)
        return UVM_ERR_INVALID_ARGUMENT;

    WARN_ON(pRecord->vma != vma); // Serious object model corruption occurred.
    WARN_ON(pRecord->baseAddress != baseAddress);
    WARN_ON(pRecord->length != length);
    WARN_ON(vma->vm_pgoff << PAGE_SHIFT != pRecord->baseAddress);

    end = pRecord->baseAddress + pRecord->length;

    pMigTracker = _get_mig_tracker(pRecord);
    if (pMigTracker == NULL)
    {
        return UVM_ERR_GPU_DMA_NOT_INITIALIZED;
    }

    WARN_ON(NULL == pRecord->osPrivate);
    WARN_ON(NULL == pRecord->osPrivate->privFile);
    WARN_ON(NULL == pRecord->osPrivate->privFile->f_mapping);

    if (pRecord->length > 0)
    {
        unmap_mapping_range(pRecord->osPrivate->privFile->f_mapping,
                            pRecord->baseAddress, pRecord->length, 1);
    }
    //
    // Copy one page at a time, from CPU to GPU. This is necessary because the
    // Copy Engine requires a physical CPU address. That in turn means that we
    // can only do one page at a time, unless we have physically contiguous
    // pages, and those are generally too expensive and risky to try to get.
    //
    for (pageVirtualAddr = pRecord->baseAddress; pageVirtualAddr < end;
         pageVirtualAddr += PAGE_SIZE, ++pageIndex)
    {
        pTracking = pRecord->commitRecordPages[pageIndex];
        if (!pTracking)
        {
            // the page is not resident on the CPU, so it doesn't get migrated:
            continue;
        }

        cpuPhysAddr = page_to_phys(pTracking->uvmPage);

        pr_debug_ratelimited("%s(pid:%u): address: 0x%p, 0x%p, vma: 0x%p\n",
                             __FUNCTION__, current->pid,
                             (void*)pageVirtualAddr,
                             (void*) cpuPhysAddr, pRecord->vma);
        //
        // TODO: handle vma splitting, mismatches between what the vma and
        // pRecord have for baseAddress and length
        //
        if (PageDirty(pTracking->uvmPage))
        {
            uvmStatus = migrate_cpu_to_gpu(pMigTracker,
                                           cpuPhysAddr,
                                           (UvmGpuPointer) pageVirtualAddr,
                                           PAGE_SIZE);
            if (uvmStatus == UVM_OK)
                migratedPages++;
        }
        //
        // Mark the page as no longer resident on the CPU, by removing its
        // pointer from the array:
        //
        pRecord->commitRecordPages[pageIndex] = NULL;

        uvm_page_cache_free_page(pTracking, pRecord->osPrivate, __FUNCTION__);

        if (uvmStatus != UVM_OK)
        {
            pr_debug_ratelimited("%s(pid:%u): failed to copy from "
                                 "cpu to gpu:"
                                 " 0x%p, 0x%p, vma: 0x%p, uvmStatus: "
                                 "0x%0x\n",
                                 __FUNCTION__, current->pid,
                                 (void*)pageVirtualAddr,
                                 (void*) cpuPhysAddr,
                                 pRecord->vma, uvmStatus);

            // Do not attempt to continue the copy operations:
            break;
        }
    }

    WARN_ON(pRecord->cachedPerProcessPerGpuCounterIndex ==
                                 UVM_INVALID_PER_PROCESS_PER_GPU_COUNTER_INDEX);

    procPerGpuCounter =
        &pRecord->osPrivate->processRecord.pProcessCounterInfo->
              procSingleGpuCounterArray[
                        pRecord->cachedPerProcessPerGpuCounterIndex];

    UvmIncrementProcessCounters(procPerGpuCounter,
        &pRecord->osPrivate->processRecord, UvmCounterNameBytesXferHtD,
        migratedPages * PAGE_SIZE);

    return uvmStatus;
}

//
// SetStreamRunning (cuda kernel launch) steps:
//
//    For each region attached to the stream ID, or to the all-stream:
//
//        1. unmap page range from user space
//
//        2. copy cpu to gpu, for dirty pages only (lots of room for
//        optimization here)
//
//        3. Free pages from page cache
//
// Locking: you must hold a write lock on the DriverPrivate.uvmPrivLock
// and mmap_sem, before calling this routine.
//
UVM_STATUS uvmlite_set_stream_running(DriverPrivate* pPriv, UvmStream streamId)
{
    struct list_head * pos;
    UvmCommitRecord * pRecord;
    UvmStreamRecord * pStream;
    UVM_STATUS uvmStatus = UVM_OK;
    UvmProcessRecord *processRecord = &pPriv->processRecord;

    pr_debug_ratelimited("%s(pid:%u): stream %lld\n",
                         __FUNCTION__, current->pid, streamId);

    if (streamId == UVM_STREAM_NONE)
    {
        // None stream should never be set to running
        WARN_ON(1);
        return UVM_ERR_INVALID_ARGUMENT;
    }

    // This might be the first time we see this streamId
    pStream = _stream_find_or_alloc(processRecord, streamId);

    if (pStream->isRunning)
    {
        // Stream is in running state already
        return UVM_OK;
    }

    list_for_each(pos, &pStream->commitRecordsList)
    {
        pRecord = list_entry(pos, UvmCommitRecord, streamRegionsListNode);
        pr_debug_ratelimited("%s(pid:%u): committed region baseAddr: 0x%p,"
                             "len: 0x%llx\n",
                             __FUNCTION__, current->pid,
                             (void*)pRecord->baseAddress,
                             pRecord->length);

        if (_is_record_matching_vma(pRecord) == UVM_FALSE)
        {
            //
            // The VMA has been modified since the record was created, skip it.
            // This should never be possible as we destroy the records with
            // modified vmas in mmap_open().
            //
            WARN_ON(1);
            continue;
        }

        // Mark the record as inaccessible.
        _set_record_inaccessible(pRecord);

        uvmStatus = uvmlite_migrate_to_gpu(pRecord->baseAddress,
                                           pRecord->length,
                                           0, // flags
                                           pRecord->vma,
                                           pRecord);
        if (uvmStatus != UVM_OK)
            goto done;
    }

    if (streamId != UVM_STREAM_ALL)
    {
        // Increment the running streams count
        ++processRecord->runningStreams;
        if (processRecord->runningStreams == 1)
        {
            // First stream to be started needs to also start the all stream
            uvmStatus = uvmlite_set_stream_running(pPriv, UVM_STREAM_ALL);
            if (uvmStatus != UVM_OK)
                goto done;
        }
    }

    pStream->isRunning = UVM_TRUE;

done:
    return uvmStatus;
}

static UVM_STATUS uvmlite_set_stream_stopped(DriverPrivate *pPriv,
    UvmStream streamId)
{
    struct list_head * pos;
    UvmCommitRecord * pRecord;
    UvmStreamRecord * pStream;
    UVM_STATUS uvmStatus = UVM_OK;
    UvmProcessRecord *processRecord = &pPriv->processRecord;

    pr_debug_ratelimited("%s(pid:%u): stream %lld\n",
                         __FUNCTION__, current->pid, streamId);

    if (streamId == UVM_STREAM_NONE)
    {
        // None stream should never be running in the first place
        WARN_ON(1);
        return UVM_ERR_INVALID_ARGUMENT;
    }

    pStream = _stream_find(processRecord, streamId);
    if (!pStream)
    {
        // The stream has never been started
        WARN_ON(1);
        return UVM_ERR_INVALID_ARGUMENT;
    }

    _stream_save_in_cache(pStream);

    if (!pStream->isRunning)
    {
        // Stream is in stopped state already
        return UVM_OK;
    }

    list_for_each(pos, &pStream->commitRecordsList)
    {
        pRecord = list_entry(pos, UvmCommitRecord, streamRegionsListNode);
        pr_debug_ratelimited("%s(pid:%u): committed region baseAddr: 0x%p,"
                             "len: 0x%llx\n",
                             __FUNCTION__, current->pid,
                             (void*)pRecord->baseAddress,
                             pRecord->length);

        if (_is_record_matching_vma(pRecord) == UVM_FALSE)
        {
            //
            // The VMA has been modified since the record was created, skip it.
            // This should never be possible as we destroy the records with
            // modified vmas in mmap_open().
            //
            WARN_ON(1);
            continue;
        }

        // Mark the record as inaccessible.
        _set_record_accessible(pRecord);
    }

    if (streamId != UVM_STREAM_ALL)
    {
        // Decrement the running stream count
        --processRecord->runningStreams;
        if (processRecord->runningStreams == 0)
        {
            // Last stream to be stopped needs to also stop the all stream
            uvmStatus = uvmlite_set_stream_stopped(pPriv, UVM_STREAM_ALL);
            if (uvmStatus != UVM_OK)
                goto done;
        }
    }

    pStream->isRunning = UVM_FALSE;

    _stream_destroy_if_empty(pStream);

done:
    return uvmStatus;
}


//
// Locking: you must hold a write lock on the DriverPrivate.uvmPrivLock
// and mmap_sem, before calling this routine.
//
UVM_STATUS uvmlite_set_streams_stopped(DriverPrivate* pPriv,
                                      UvmStream * streamIdArray,
                                      UvmSize nStreams)
{
    UvmSize i;
    UVM_STATUS uvmStatus = UVM_OK;

    for (i = 0; i < nStreams; ++i)
    {
        uvmStatus = uvmlite_set_stream_stopped(pPriv, streamIdArray[i]);
        if (uvmStatus != UVM_OK)
            break;
    }

    return uvmStatus;
}

static void _stream_destroy_if_empty(UvmStreamRecord *pStream)
{
    if (!pStream)
        return;

    if (pStream->isRunning)
    {
        //
        // Don't destroy running streams even if they are empty as a record
        // might be attached before they are stopped.
        //
        return;
    }

    if (!list_empty(&pStream->commitRecordsList))
    {
        // Don't destroy streams with attached records
        return;
    }

    _stream_destroy(pStream);
}

UVM_STATUS uvmlite_region_set_stream(UvmCommitRecord *pRecord,
    UvmStream newStreamId)
{
    UvmProcessRecord *pProccesRecord = &pRecord->osPrivate->processRecord;
    UvmStreamRecord *pNewStream;
    UVM_STATUS status;

    if (pRecord == NULL)
        return UVM_ERR_INVALID_ARGUMENT;

    if (pRecord->pStream->streamId == newStreamId)
        return UVM_OK;

    pNewStream = _stream_find_or_alloc(pProccesRecord, newStreamId);
    if (!pNewStream)
        return UVM_ERR_INSUFFICIENT_RESOURCES;

    status = _record_attach_to_stream(pRecord, pNewStream);

    return status;
}

void uvmlite_free_process_counters(UvmProcessCounterInfo *pProcCounterInfo)
{
    unsigned gpu;
    UvmCounterInfo *pCurrCounter;

    for (gpu = 0; gpu < UVM_MAX_GPUS; gpu++)
    {
        pCurrCounter = &pProcCounterInfo->procSingleGpuCounterArray[gpu];

        if(pCurrCounter->pCounterPage != NULL)
        {
            kunmap(pCurrCounter->pCounterPage);
            __free_page(pCurrCounter->pCounterPage);
        }
    }

    pCurrCounter = &pProcCounterInfo->procAllGpuCounter;

    if(pCurrCounter->pCounterPage != NULL)
    {
        kunmap(pCurrCounter->pCounterPage);
        __free_page(pCurrCounter->pCounterPage);
    }
    kmem_cache_free(g_uvmProcessCounterInfoCache, pProcCounterInfo);
}

//
// This function initializes the counter information structure pCtrInfo.
//
static UVM_STATUS _create_counter
(
    UvmCounterInfo *pCtrInfo
)
{
    memset(pCtrInfo, 0, sizeof(*pCtrInfo));
    pCtrInfo->pCounterPage = alloc_page(GFP_KERNEL | GFP_HIGHUSER);

    if (!pCtrInfo->pCounterPage)
        return UVM_ERR_INSUFFICIENT_RESOURCES;

    pCtrInfo->sysAddr = kmap(pCtrInfo->pCounterPage);
    memset(pCtrInfo->sysAddr, 0, PAGE_SIZE);

    return UVM_OK;
}

UVM_STATUS uvmlite_init_process_counters
(
    UvmProcessRecord *pRecord
)
{
    int gpu;
    UvmProcessCounterInfo* pProcessCounterInfo;

    UVM_STATUS status;

    pProcessCounterInfo = (UvmProcessCounterInfo *)
        kmem_cache_alloc(g_uvmProcessCounterInfoCache, GFP_KERNEL);

    if(pProcessCounterInfo == NULL)
        return UVM_ERR_INSUFFICIENT_RESOURCES;

    memset(pProcessCounterInfo, 0, sizeof(UvmProcessCounterInfo));
    NV_ATOMIC_SET(pProcessCounterInfo->refcountUsers, 1);

    for (gpu = 0; gpu < UVM_MAX_GPUS; gpu++)
    {
        status = _create_counter(&pProcessCounterInfo->
                    procSingleGpuCounterArray[gpu]);
        if (status != UVM_OK)
            goto fail;
    }

    status = _create_counter(&pProcessCounterInfo->procAllGpuCounter);
    if(status != UVM_OK)
        goto fail;


    init_rwsem(&pProcessCounterInfo->gpuHistoryLock);
    memset(pProcessCounterInfo->gpuHistory, 0,
        sizeof(pProcessCounterInfo->gpuHistory));
    pProcessCounterInfo->numGpus = 0;
    pRecord->pProcessCounterInfo = pProcessCounterInfo;

    return UVM_OK;

fail:
    // free all the resources that got allocated before failure
    uvmlite_free_process_counters(pProcessCounterInfo);
    return status;

}


//
// Locking: this routine initializes the DriverPrivate.uvmPrivLock.
// thread safety: uvmlite_open has to be used by a single thread,
//
static int uvmlite_open(struct inode *inode, struct file *filp)
{
    DriverPrivate * pUvmPrivate = NULL;
    int retval = -ENOMEM;

    struct address_space *pMapping =
        (struct address_space*)kmem_cache_alloc(g_uvmMappingCache, GFP_KERNEL);

    if (!pMapping)
        goto fail;

    pUvmPrivate = (DriverPrivate*)
                  kmem_cache_alloc(g_uvmPrivateCache, GFP_KERNEL);

    if (!pUvmPrivate)
        goto fail;
    //
    // UVM-Lite calls unmap_mapping_range, but UVM-Lite has only a single device
    // node, through which all user-space processes do their mmap() calls. In
    // order to avoid interference among unrelated processes, we must set up a
    // separate mapping object for each struct file.
    //
    address_space_init_once(pMapping);
    pMapping->host = inode;
    filp->f_mapping = pMapping;

    memset(pUvmPrivate, 0, sizeof(DriverPrivate));
    INIT_LIST_HEAD(&pUvmPrivate->pageList);
    INIT_LIST_HEAD(&pUvmPrivate->processRecord.allStreamList);
    INIT_LIST_HEAD(&pUvmPrivate->processRecord.processRecordNode);
    init_rwsem(&pUvmPrivate->uvmPrivLock);
    pUvmPrivate->privFile = filp;
    pUvmPrivate->processRecord.euid = NV_CURRENT_EUID();

    if (UVM_OK != uvmlite_init_process_counters(&pUvmPrivate->processRecord))
       goto fail;

    // Initialize session information array
    NvUvmInitSessionInfoArray(&pUvmPrivate->processRecord);

    filp->private_data = pUvmPrivate;

    pUvmPrivate->processRecord.pid = UvmGetProcessId();
    // register to global process record table after initialization
    down_write(&g_uvmProcessRecordTableLock);
    list_add(&pUvmPrivate->processRecord.processRecordNode, 
             &g_uvmProcessRecordList); 
    up_write(&g_uvmProcessRecordTableLock);

    pr_devel("%s(pid:%u): pPriv: 0x%p, f_mapping: 0x%p\n",
             __FUNCTION__, current->pid, filp->private_data,
             filp->f_mapping);

    return 0;

fail:
    if (!pMapping)
        kmem_cache_free(g_uvmMappingCache, pMapping);

    if (!pUvmPrivate)
        kmem_cache_free(g_uvmPrivateCache, pUvmPrivate);

    return retval;
}

UvmProcessCounterInfo *
uvmlite_get_and_refcount_process_counter_info(UvmProcessRecord *pProcessRecord)
{
    UvmProcessCounterInfo *pProcessCounterInfo = 
                                pProcessRecord->pProcessCounterInfo;

    NV_ATOMIC_INC(pProcessCounterInfo->refcountUsers);

    return pProcessCounterInfo;
}


void uvmlite_put_and_unrefcount_process_counter_info
(
    UvmProcessCounterInfo *pProcessCounterInfo
)
{
    // Free counter information if this is the only process using it.
    if (NV_ATOMIC_DEC_AND_TEST(pProcessCounterInfo->refcountUsers))
        uvmlite_free_process_counters(pProcessCounterInfo);
}

//
// Locking: you must hold processRecordTableLock read lock 
// when you access UvmProcessRecord.
// 
static UvmProcessRecord* _find_process_record(unsigned pid)
{
    struct list_head * pos;
    UvmProcessRecord *pProcessRecord;

    list_for_each(pos, &g_uvmProcessRecordList)
    {
        pProcessRecord = list_entry(pos, UvmProcessRecord, 
                                    processRecordNode);
        if (pid == pProcessRecord->pid)
            return pProcessRecord;
    }
    return NULL;
}


//
// On success, this routine increments refcount on UmvProcessCouterInfo
// and returns pEuid and UvmProcessCounterInfo.
// Locking:
//     1. This routine acquires the read g_uvmProcessRecordTableLock.
//
UVM_STATUS uvmlite_secure_get_process_counter_info
(
    unsigned pidTarget,
    UvmProcessCounterInfo **ppProcessCounterInfo,
    uid_t *pEuid
)
{
    UvmProcessRecord *pProcRec;


    //
    // uvmlite_close can't decrement refcount / remove ProcessCounterInfo
    // structure before grabbing g_uvmProcessRecordTableLock and removing itself
    // from g_uvmProcessRecordList.
    //
    down_read(&g_uvmProcessRecordTableLock);

    pProcRec = _find_process_record(pidTarget);

    if (pProcRec == NULL)
    {
        up_read(&g_uvmProcessRecordTableLock);
        return UVM_ERR_INVALID_ARGUMENT;
    }

    *pEuid = pProcRec->euid;

    if (!UvmSecurityCheck(*pEuid))
    {
        up_read(&g_uvmProcessRecordTableLock);
        return UVM_ERR_NOT_PERMITTED;
    }

    *ppProcessCounterInfo = 
        uvmlite_get_and_refcount_process_counter_info(pProcRec);

    up_read(&g_uvmProcessRecordTableLock);

    return UVM_OK;
}

//
// Locking:
//     1. This routine acquires the g_uvmProcessRecordTableLock.
//
static int uvmlite_close(struct inode *inode, struct file *filp)
{
    DriverPrivate* pPriv = (DriverPrivate*)filp->private_data;

    // unregister from global process record table before cleanup
    down_write(&g_uvmProcessRecordTableLock);
    list_del(&pPriv->processRecord.processRecordNode);
    up_write(&g_uvmProcessRecordTableLock);

    uvmlite_put_and_unrefcount_process_counter_info(
        pPriv->processRecord.pProcessCounterInfo);
    //
    // At this point all the regions have been removed, but there might be some
    // leftover streams in running state.
    //
    _stop_and_destroy_leftover_streams(&pPriv->processRecord);
    WARN_ON(!list_empty(&pPriv->processRecord.allStreamList));

    _delete_all_migration_resources(pPriv);
    uvm_page_cache_free_for_this_process(pPriv, __FUNCTION__);
    _delete_all_session_info_table_entries(&pPriv->processRecord);

    kmem_cache_free(g_uvmMappingCache, filp->f_mapping);
    kmem_cache_free(g_uvmPrivateCache, pPriv);
    pr_devel("%s(pid:%u): done\n", __FUNCTION__, current->pid);

    return 0;
}

static int uvmlite_mmap(struct file * filp, struct vm_area_struct * vma)
{
    UvmCommitRecord * pRecord = (UvmCommitRecord*)vma->vm_private_data;

    if (pRecord)
    {
        pr_debug_ratelimited("mmap(pid:%u): f_mapping: 0x%p, vma 0x%p [0x%p, 0x%p)\n",
                 current->pid, pRecord->osPrivate->privFile->f_mapping,
                 vma, (void*)vma->vm_start, (void*)vma->vm_end);
    }

    vma->vm_ops   = &uvmlite_vma_ops;

    // Prohibit copying the vma on fork().
    vma->vm_flags |= VM_DONTCOPY;
    // Prohibt mremap() that would expand the vma.
    vma->vm_flags |= VM_DONTEXPAND;

    // Other cases of vma modification are detected in _mmap_open().

    //
    // No access until we know which stream the vma (region) is attached to.
    // This will be done in the REGION_COMMIT ioctl.
    //
    _set_vma_inaccessible(vma);

    return 0;
}

static long uvmlite_unlocked_ioctl(struct file *filp,
                                   unsigned int cmd,
                                   unsigned long arg)
{
    //
    // The following macro is only intended for use in this routine. That's why
    // it is declared inside the function (even though, of course, the
    // preprocessor ignores such scoping).
    //
    #define UVM_ROUTE_CMD(cmd,functionName)                                 \
        case cmd:                                                           \
        {                                                                   \
            cmd##_PARAMS params;                                            \
            if (copy_from_user(&params, (void __user*)arg, sizeof(params))) \
                return -EFAULT;                                             \
                                                                            \
            params.uvmStatus = functionName(&params, filp);                 \
            if (copy_to_user((void __user*)arg, &params, sizeof(params)))   \
                return -EFAULT;                                             \
                                                                            \
            break;                                                          \
        }

    switch (cmd)
    {
        case UVM_INITIALIZE:
            pr_devel("%s(pid:%u): cmd: UVM_INITIALIZE\n", __FUNCTION__,
                     current->pid);
            break;
        case UVM_DEINITIALIZE:
            pr_devel("%s(pid:%u): cmd: UVM_DEINITIALIZE\n", __FUNCTION__,
                     current->pid);
            break;
        UVM_ROUTE_CMD(UVM_RESERVE_VA,             uvmAPI_ReserveVa);
        UVM_ROUTE_CMD(UVM_RELEASE_VA,             uvmAPI_ReleaseVa);
        UVM_ROUTE_CMD(UVM_REGION_COMMIT,          uvmAPI_RegionCommit);
        UVM_ROUTE_CMD(UVM_REGION_DECOMMIT,        uvmAPI_RegionDecommit);
        UVM_ROUTE_CMD(UVM_REGION_SET_STREAM,      uvmAPI_RegionSetStream);
        UVM_ROUTE_CMD(UVM_SET_STREAM_RUNNING,     uvmAPI_SetStreamRunning);
        UVM_ROUTE_CMD(UVM_SET_STREAM_STOPPED,     uvmAPI_SetStreamStopped);
        UVM_ROUTE_CMD(UVM_MIGRATE_TO_GPU,         uvmAPI_MigrateToGpu);
#if defined(DEBUG)
        UVM_ROUTE_CMD(UVM_RUN_TEST,               uvmAPI_RunTest);
#endif
        UVM_ROUTE_CMD(UVM_ADD_SESSION,            uvmAPI_AddSession);
        UVM_ROUTE_CMD(UVM_REMOVE_SESSION,         uvmAPI_RemoveSession);
        UVM_ROUTE_CMD(UVM_MAP_COUNTER,            uvmAPI_MapCounter);
        UVM_ROUTE_CMD(UVM_ENABLE_COUNTERS,        uvmAPI_EnableCounters);
        default:
            pr_err("%s(pid:%u): Unknown: cmd: 0x%0x\n",
                   __FUNCTION__, current->pid, cmd);
            return -EINVAL;
            break;
    }

    #undef UVM_ROUTE_CMD

    return 0;
}

static const struct file_operations uvmlite_fops = {
    .open            = uvmlite_open,
    .release         = uvmlite_close,
    .mmap            = uvmlite_mmap,
    .unlocked_ioctl  = uvmlite_unlocked_ioctl,
#if defined(__X86_64__) && defined(NV_FILE_OPERATIONS_HAS_COMPAT_IOCTL)
    .compat_ioctl    = uvmlite_unlocked_ioctl,
#endif
    .owner           = THIS_MODULE,
};


//
// Locking: this initializes the g_uvmProcessRecordTableLock, and doesn't take or acquire
// any other locks.
//
int uvmlite_init(dev_t uvmBaseDev)
{
    dev_t uvmliteDev = MKDEV(MAJOR(uvmBaseDev), NVIDIA_UVM_LITE_MINOR_NUMBER);
    int ret = 0;
    int cdevAlloced = 0;

    cdev_init(&g_uvmlite_cdev, &uvmlite_fops);
    g_uvmlite_cdev.owner = THIS_MODULE;

    ret = cdev_add(&g_uvmlite_cdev, uvmliteDev, 1);
    if (ret)
    {
        pr_err("%s: cdev_add (major %u, minor %u) failed: %d\n",
               __FUNCTION__, ret, MAJOR(uvmliteDev), MINOR(uvmliteDev));
        goto fail;
    }
    cdevAlloced = 1;

    // TODO: call register_shrinker()
    init_rwsem(&g_uvmProcessRecordTableLock);

    // Debugging hint: kmem_cache_create objects are visible in /proc/slabinfo:
    ret = -ENOMEM;
    NV_KMEM_CACHE_CREATE(g_uvmPrivateCache, "uvm_private_t",
                         struct DriverPrivate_tag);
    if (!g_uvmPrivateCache)
        goto fail;

    NV_KMEM_CACHE_CREATE(g_uvmCommitRecordCache, "uvm_commit_record_t",
                         struct UvmCommitRecord_tag);
    if (!g_uvmCommitRecordCache)
        goto fail;

    NV_KMEM_CACHE_CREATE(g_uvmProcessCounterInfoCache,
        "uvm_process_counter_info_t", struct UvmProcessCounterInfo_tag);

    if(!g_uvmProcessCounterInfoCache)
        goto fail;

    NV_KMEM_CACHE_CREATE(g_uvmMigTrackerCache, "uvm_mig_tracker_t",
                         struct UvmGpuMigrationTracking_tag);

    if (!g_uvmMigTrackerCache)
        goto fail;

    NV_KMEM_CACHE_CREATE(g_uvmStreamRecordCache, "uvm_stream_record_t",
                         struct UvmStreamRecord_tag);
    if (!g_uvmStreamRecordCache)
        goto fail;

    NV_KMEM_CACHE_CREATE(g_uvmMappingCache, "uvm_mapping_t",
                         struct address_space);
    if (!g_uvmMappingCache)
        goto fail;

    if (0 != uvm_page_cache_init())
        goto fail;

    return 0;

fail:
    kmem_cache_destroy_safe(g_uvmMappingCache);
    kmem_cache_destroy_safe(g_uvmStreamRecordCache);
    kmem_cache_destroy_safe(g_uvmMigTrackerCache);
    kmem_cache_destroy_safe(g_uvmProcessCounterInfoCache);
    kmem_cache_destroy_safe(g_uvmCommitRecordCache);
    kmem_cache_destroy_safe(g_uvmPrivateCache);

    if (cdevAlloced)
        cdev_del(&g_uvmlite_cdev);

    pr_err("%s: Failed\n", __FUNCTION__);
    return ret;

}

int uvmlite_setup_gpu_list()
{
    NvU8 *pUuidList;
    UVM_STATUS status;
    unsigned i = 0;
    UvmGpuProviderTracking *provider;  
    unsigned numAttachedGpus = 0;
  
    // get the list of gpus
    provider = getProvider();
    if(!provider)
       return -ENODEV;
   
    // get the list of gpus
    pUuidList = vmalloc(sizeof(NvU8) * UVM_MAX_GPUS * UVM_UUID_LEN);
    if (!pUuidList)
        return -ENOMEM;

    status = provider->registeredGpuOps->getAttachedUuids(pUuidList, 
                                                         &numAttachedGpus);
    if (status != UVM_OK || (numAttachedGpus > UVM_MAX_GPUS))
    {
        pr_err("%s: ERROR:Error in finding GPU's\n",__FUNCTION__);
        vfree(pUuidList);
        return -ENODEV;
    }

    // construct the uuid list 
    for (i = 0; i < numAttachedGpus; i++)
    {
        memcpy(&g_attached_uuid_list[i].uuid[0], 
               (pUuidList + (i * UVM_UUID_LEN)), UVM_UUID_LEN);    
        uvm_print_uuid(__FUNCTION__, "Found attached GPU",
                       &g_attached_uuid_list[i]);
    }
    vfree(pUuidList);
    return 0;
}

void uvmlite_exit(void)
{
    //
    // No extra cleanup of regions or data structures is necessary here, because
    // that is done by the file release routine, and the kernel won't allow the
    // module to be unloaded while it's device file open count remains >0.
    //
    // However, this is still a good place for:
    //
    //     TODO: check for resource leaks here, just in case.
    //

    // TODO: call unregister_shrinker()

    kmem_cache_destroy(g_uvmMappingCache);
    kmem_cache_destroy(g_uvmStreamRecordCache);
    kmem_cache_destroy(g_uvmMigTrackerCache);
    kmem_cache_destroy(g_uvmCommitRecordCache);
    kmem_cache_destroy(g_uvmPrivateCache);
    kmem_cache_destroy(g_uvmProcessCounterInfoCache);

    uvm_page_cache_destroy();

    cdev_del(&g_uvmlite_cdev);
}

//
// Locking: you must hold a write lock on the DriverPrivate.uvmPrivLock, before
// calling this routine.
//
static int _create_migration_resources(UvmGpuUuid * pGpuUuidStruct,
                                       UvmGpuMigrationTracking *pMigTracking)
{
    // this function will use the ops to setup the copy engine and the channel

    UVM_STATUS status;
    UvmGpuProviderTracking *provider;
    unsigned ceInstance = 1;
    unsigned long long uuidMsb = 0;
    unsigned long long uuidLsb = 0;

    if(!pMigTracking)
        return -EINVAL;

    uvm_print_uuid(__FUNCTION__, "Entering", pGpuUuidStruct);

    provider = getProvider();

    if(!provider)
        return -ENODEV;

    status = provider->registeredGpuOps->sessionCreate(&pMigTracking->hSession);
    if(status != UVM_OK)
    {
        pr_err("%s(pid:%u): ERROR: could not create a session\n",
               __FUNCTION__, current->pid);
        goto cleanup;
    }

    memcpy(&uuidMsb, &pGpuUuidStruct->uuid[0], (UVM_UUID_LEN >> 1));
    memcpy(&uuidLsb, &pGpuUuidStruct->uuid[8], (UVM_UUID_LEN >> 1));
    status = provider->registeredGpuOps->addressSpaceCreateMirrored(
                               pMigTracking->hSession,
                               uuidMsb,
                               uuidLsb,
                               &pMigTracking->hVaSpace);
    if(status != UVM_OK)
    {
        pr_err("%s(pid:%u): ERROR: could not create an address space\n",
               __FUNCTION__, current->pid);
        goto cleanup_session;
    }

    status = provider->registeredGpuOps->channelAllocate(
                                             pMigTracking->hVaSpace,
                                             &pMigTracking->hChannel,
                                             &pMigTracking->channelInfo);
    if(status != UVM_OK)
    {
        pr_err("%s(pid:%u): ERROR: could not allocate a channel\n",
               __FUNCTION__, current->pid);
        goto cleanup_addressspace;
    }

    status = UVM_ERR_FAILED;

    for (ceInstance = 1; ceInstance <= MAX_NUM_COPY_ENGINES; ++ceInstance)
    {
        status = provider->registeredGpuOps->copyEngineAllocate(
                                              pMigTracking->hChannel,
                                              ceInstance,
                                              &pMigTracking->ceClassNumber,
                                              &pMigTracking->hCopyEngine);

        if ((status == UVM_ERR_INVALID_INDEX) || (status == UVM_OK))
            break;
    }

    if (status != UVM_OK)
    {
        pr_err("%s(pid:%u): ERROR: could not allocate OBJCE\n",
               __FUNCTION__, current->pid);
       goto cleanup_addressspace;
    }

    // allocate semaphore page
    status = provider->registeredGpuOps->allocGpuMemorySys(
                                        pMigTracking->hVaSpace, SEMAPHORE_SIZE,
                                         &pMigTracking->gpuSemaPtr);
    if (UVM_OK != status)
    {
        pr_err("%s(pid:%u): ERROR: could not allocate GPU memory for PB\n",
               __FUNCTION__, current->pid);
        goto cleanup_addressspace;
    }
    status = provider->registeredGpuOps->cpuMap(pMigTracking->hVaSpace,
                                                pMigTracking->gpuSemaPtr,
                                                SEMAPHORE_SIZE,
                                                &pMigTracking->cpuSemaPtr);
    if (UVM_OK != status)
    {
        pr_err("%s(pid:%u): ERROR: could not map PB to CPU VA\n",
               __FUNCTION__, current->pid);
        goto cleanup_addressspace;
    }

    // allocate a Push Buffer segment
    status = provider->registeredGpuOps->allocGpuMemorySys(
                                         pMigTracking->hVaSpace,
                                         PUSHBUFFER_SIZE,
                                         &pMigTracking->gpuPushBufferPtr);
    if (UVM_OK != status)
    {
        pr_err("%s(pid:%u): ERROR: could not allocate GPU memory for PB\n",
               __FUNCTION__, current->pid);
        goto cleanup_addressspace;
    }
    // Map Push Buffer
    status = provider->registeredGpuOps->cpuMap(
                                         pMigTracking->hVaSpace,
                                         pMigTracking->gpuPushBufferPtr,
                                         PUSHBUFFER_SIZE,
                                         &pMigTracking->cpuPushBufferPtr);
    if (UVM_OK != status)
    {
        pr_err("%s(pid:%u): ERROR: could not map PB to CPU VA\n",
               __FUNCTION__, current->pid);
        goto cleanup_addressspace;
    }

    // setup CE Ops
    status = NvUvmHalInit(pMigTracking->ceClassNumber,
                 pMigTracking->channelInfo.channelClassNum,
                 &pMigTracking->ceOps);

    if (UVM_OK != status)
    {
        pr_err("%s(pid:%u): ERROR: could not find a CE HAL to use\n",
               __FUNCTION__, current->pid);
        goto cleanup_addressspace;
    }

    pr_devel("%s(pid:%u): Done. channelClassNum: 0x%0x, ceClassNum: 0x%0x\n",
             __FUNCTION__, current->pid,
             pMigTracking->channelInfo.channelClassNum,
             pMigTracking->ceClassNumber);


    return 0;

cleanup_addressspace:
    provider->registeredGpuOps->addressSpaceDestroy(pMigTracking->hVaSpace);
cleanup_session:
    provider->registeredGpuOps->sessionDestroy(pMigTracking->hSession);
cleanup:
    return status;
}

//
// Locking: you must hold a write lock on the DriverPrivate.uvmPrivLock, before
// calling this routine.
//
static void _destroy_migration_resources(UvmGpuMigrationTracking *pMigTracking)
{
    UvmGpuProviderTracking *provider;

    if(!pMigTracking)
        return;

    pr_devel("%s(pid:%u): Entering\n", __FUNCTION__, current->pid);
    provider = getProvider();

    if(!provider)
        return;

    // destroy the channel and the engines under it
    if (pMigTracking->hChannel != 0)
        provider->registeredGpuOps->channelDestroy(pMigTracking->hChannel);

    if (pMigTracking->hVaSpace != 0)
    {
        provider->registeredGpuOps->addressSpaceDestroy(pMigTracking->hVaSpace);
        //
        // TODO: Fix the RM bug where it thinks allocating a bogus session
        // is ok when there is no valid device.
        //
        provider->registeredGpuOps->sessionDestroy(pMigTracking->hSession);
    }

    pr_devel("%s(pid:%u): Done\n", __FUNCTION__, current->pid);

    return;
}

//
// This function will migrate pages from CPU sysmem
// to GPU video memory. The CPU pointer is physical
// and the GPU pointer is virtual.
//
static
int _migrate_data(UvmGpuMigrationTracking *pMigTracker,
                  UvmGpuPointer srcGpuPtr,
                  UvmGpuPointer destCpuPtr,
                  unsigned srcFlags,
                  unsigned destFlags,
                  unsigned launchFlags,
                  UvmSize length)
{
    UvmSize numMethods = 0;
    UvmCopyOps *pCopyOps;
    char *cpuPbPointer = NULL;
    char * cpuPbEnd = NULL;
    unsigned semaVal = 0;

    if(!pMigTracker || (pMigTracker->ceOps.launchDma == NULL) ||
       (pMigTracker->ceOps.writeGpEntry == NULL))
        return -EINVAL;

    //
    // Since it's a synchronous copy we can keep reusing the same PB region. We
    // don't need to worry about overlap, etc.
    //
    pCopyOps = &pMigTracker->ceOps;
    cpuPbPointer = pMigTracker->cpuPushBufferPtr;
    cpuPbEnd     = pMigTracker->cpuPushBufferPtr + PUSHBUFFER_SIZE;

    // reset the semaphore payload
    *(unsigned *)pMigTracker->cpuSemaPtr = 0;

    // push the methods to do the copy:
    numMethods += pCopyOps->launchDma((unsigned **)&cpuPbPointer,
                                      (unsigned *) cpuPbEnd,
                                      srcGpuPtr, srcFlags,
                                      destCpuPtr, destFlags,
                                      length,
                                      launchFlags);

    // push methods to release the semaphore
    numMethods += pCopyOps->semaphoreRelease((unsigned **)&cpuPbPointer,
                                             (unsigned*)cpuPbEnd,
                                             pMigTracker->gpuSemaPtr,
                                             0xFACEFEED);

    // wrap around gpFifoOffset if needed
    if (pMigTracker->channelInfo.numGpFifoEntries ==
            pMigTracker->currentGpFifoOffset + 1)
    {
        pMigTracker->currentGpFifoOffset = 0;
    }

    // write the GP entry
    pCopyOps->writeGpEntry(pMigTracker->channelInfo.gpFifoEntries,
                           pMigTracker->currentGpFifoOffset,
                           pMigTracker->gpuPushBufferPtr,
                           numMethods);
    // launch the copy
    NvUvmChannelWriteGpPut(pMigTracker->channelInfo.GPPut,
                           pMigTracker->currentGpFifoOffset+1);
    pMigTracker->currentGpFifoOffset++;

    //
    // spin on the semaphore before returning
    // TODO: consider busy-waiting vs. sleeping, probably depending on copy
    // size.
    //
    pr_debug_ratelimited("%s(pid:%u): starting page migration for CPU "
                         "physical-GPU Virtual address: 0x%p - 0x%p\n",
                         __FUNCTION__, current->pid,
                         (void*)destCpuPtr, (void *)srcGpuPtr);

    semaVal = 0;
    while(semaVal != 0xFACEFEED)
    {
        semaVal = *(volatile unsigned *)pMigTracker->cpuSemaPtr;
        cpu_relax();
    }

    return 0;
}

int migrate_gpu_to_cpu(UvmGpuMigrationTracking *pMigTracker,
                       UvmGpuPointer srcGpuPtr,
                       UvmGpuPointer destCpuPtr,
                       UvmSize length)
{
    int status;
    unsigned srcFlags = NV_UVM_COPY_SRC_LOCATION_FB;
    unsigned destFlags = NV_UVM_COPY_DST_LOCATION_SYSMEM;
    unsigned launchFlags = NV_UVM_COPY_DST_TYPE_PHYSICAL |
                            NV_UVM_COPY_SRC_TYPE_VIRTUAL;

    pr_debug_ratelimited("%s(pid:%u): starting. CPU physical address: 0x%p, "
                         "len: 0x%x\n",
                         __FUNCTION__, current->pid,
                         (void*)destCpuPtr, (unsigned) length);

    status = _migrate_data(pMigTracker, srcGpuPtr, destCpuPtr,
                         srcFlags, destFlags, launchFlags,
                         length);

    pr_debug_ratelimited("%s(pid:%u): done. CPU physical address: 0x%p, "
                         "len: 0x%x\n",
                         __FUNCTION__, current->pid,
                         (void*)destCpuPtr, (unsigned) length);
    return status;
}

//
// This function will migrate pages from GPU video memory
// to CPU sysmem. The CPU pointer is physical
// and the GPU pointer is virtual.
//
int migrate_cpu_to_gpu(UvmGpuMigrationTracking *pMigTracker,
                       UvmGpuPointer srcCpuPtr,
                       UvmGpuPointer destGpuPtr,
                       UvmSize length)
{
    int status;
    unsigned srcFlags = NV_UVM_COPY_SRC_LOCATION_SYSMEM;
    unsigned destFlags = NV_UVM_COPY_DST_LOCATION_FB;
    unsigned launchFlags = NV_UVM_COPY_DST_TYPE_VIRTUAL |
                            NV_UVM_COPY_SRC_TYPE_PHYSICAL;

    pr_debug_ratelimited("%s(pid:%u): starting. CPU physical address: 0x%p, "
                         "len: 0x%x\n",
             __FUNCTION__, current->pid,
             (void*)srcCpuPtr, (unsigned) length);

    status = _migrate_data(pMigTracker, srcCpuPtr, destGpuPtr,
                         srcFlags, destFlags, launchFlags,
                         length);

    pr_debug_ratelimited("%s(pid:%u): done. CPU physical address: 0x%p, "
                         "len: 0x%x\n",
             __FUNCTION__, current->pid,
             (void*)srcCpuPtr, (unsigned) length);
    return status;
}

//
// Locking: you must hold a write lock on the mmap_sem.
//
static void _set_vma_inaccessible(struct vm_area_struct * vma)
{
    //
    // Disable VM_MAY* to disallow mprotect() from changing the perms.
    // Subsequent access from userspace after the pages are unmapped will cause
    // a SIGSEGV.
    //
    vma->vm_flags &= ~(VM_READ|VM_MAYREAD);
    vma->vm_flags &= ~(VM_WRITE|VM_MAYWRITE);
}

//
// Locking: you must hold a write lock on the mmap_sem.
//
static void _set_vma_accessible(struct vm_area_struct * vma)
{
    vma->vm_flags |= (VM_READ|VM_MAYREAD);
    vma->vm_flags |= (VM_WRITE|VM_MAYWRITE);
}

//
// Locking: you must hold a write lock on the mmap_sem.
//
static void _set_record_accessible(UvmCommitRecord *pRecord)
{
    _set_vma_accessible(pRecord->vma);
}

//
// Locking: you must hold a write lock on the mmap_sem.
//
static void _set_record_inaccessible(UvmCommitRecord *pRecord)
{
    _set_vma_inaccessible(pRecord->vma);
}

//
// Locking: you must hold a read lock on the mmap_sem
// and DriverPrivate.uvmPrivLock.
//
static int _is_record_matching_vma(UvmCommitRecord *pRecord)
{
    if (pRecord->baseAddress != pRecord->vma->vm_start
        || PAGE_ALIGN(pRecord->length) != 
        (pRecord->vma->vm_end - pRecord->vma->vm_start))
    {
        return UVM_FALSE;
    }
    return UVM_TRUE;
}
