//TODO: 1368768: follow up work

/*******************************************************************************
    Copyright (c) 2013 NVidia Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

//
// This file contains supporting functions to get proper process/ gpu 
// information required for incrementing/ enabling/ mapping counters.
//
#include "nvidia_uvm_lite_counters.h"
#include "nvidia_uvm_lite.h"
#include "uvm_ioctl.h"

// Session index is initialized to an invalid value.
static const int UVM_SESSION_INDEX_INIT_VALUE   = -1;
// number of counters that can fit in a page.
static const int UVM_MAX_COUNTERS_PER_PAGE  = 
                            PAGE_SIZE / sizeof (unsigned long long);
// shift for per process per gpu counters in user mapping.
static const int UVM_PER_PROCESS_PER_GPU_COUNTERS_SHIFT =
                            PAGE_SIZE / sizeof (unsigned long long);

void NvUvmInitSessionInfoArray(UvmProcessRecord *pCurrentProcessRecord)
{
    init_rwsem(&pCurrentProcessRecord->sessionInfoLock);
    memset(pCurrentProcessRecord->sessionInfoArray, 0, 
           sizeof (pCurrentProcessRecord->sessionInfoArray));
}

// sessionInfoLock lock must be acquired before calling this function
static UVM_STATUS NvUvmGetSessionInfo
(
    int sessionIndex, 
    UvmProcessRecord *pCurrentProcessRecord,
    UvmSessionInfo **pSessionInfo
)
{
    // check if session index is valid
    if((sessionIndex > UVM_SESSION_INDEX_INIT_VALUE) &&
      (sessionIndex < UVM_MAX_SESSIONS_PER_PROCESS))
    {
        unsigned pidCurrent = UvmGetProcessId();

        // check if the current process id is the session owner
        if (pCurrentProcessRecord->sessionInfoArray[sessionIndex].
            pidSessionOwner != pidCurrent)
        {
            return UVM_ERR_NOT_PERMITTED;
        }

        *pSessionInfo = &pCurrentProcessRecord->sessionInfoArray[sessionIndex];

        return UVM_OK;
    }

    return UVM_ERR_INVALID_ARGUMENT;
}


//
// Locking: this function uses sessionInfoLock
// 
UVM_STATUS NvUvmAddSessionInfo
(
    unsigned pidTarget,
    uid_t euidTarget,
    int *pSessionIndex,
    UvmProcessCounterInfo *pProcessCounterInfo,
    unsigned long long *mappedUserBaseAddress,
    UvmProcessRecord *pCurrentProcessRecord
)
{
    int sessionArrayIndex;
    int i;
    UvmSessionInfo *pSessionInfo = NULL;
    UVM_STATUS status = UVM_ERR_INSUFFICIENT_RESOURCES;
    unsigned long long *currentUserBaseAddress;

    down_write(&pCurrentProcessRecord->sessionInfoLock);

    for(sessionArrayIndex = 0; sessionArrayIndex < UVM_MAX_SESSIONS_PER_PROCESS;
        sessionArrayIndex++)
    {
        if (pCurrentProcessRecord->sessionInfoArray[sessionArrayIndex].
            pidSessionOwner ==  UVM_PID_INIT_VALUE)
        {
            pSessionInfo = 
                &pCurrentProcessRecord->sessionInfoArray[sessionArrayIndex];

            // save a pointer to the counter information of the target process
            pSessionInfo->pTargetCounterInfo = pProcessCounterInfo;
            // save the owner pid (used later for validating the session owner)
            pSessionInfo->pidSessionOwner = UvmGetProcessId();
            pSessionInfo->pidTarget = pidTarget;
            pSessionInfo->euidTarget = euidTarget;

            *pSessionIndex = sessionArrayIndex;
            status = UVM_OK;
            break;
        }
    }
    up_write(&pCurrentProcessRecord->sessionInfoLock);

    if (pSessionInfo == NULL)
    {
        return UVM_ERR_INSUFFICIENT_RESOURCES;
    }
    
    currentUserBaseAddress = mappedUserBaseAddress;
    status = UvmMapPage(pProcessCounterInfo->procAllGpuCounter.pCounterPage, 
                         currentUserBaseAddress);
    currentUserBaseAddress += UVM_MAX_COUNTERS_PER_PAGE;
    for (i = 0; i < UVM_MAX_GPUS; ++i)
    {
        UvmMapPage(pProcessCounterInfo->procSingleGpuCounterArray[i].
                    pCounterPage, currentUserBaseAddress);
        currentUserBaseAddress += UVM_MAX_COUNTERS_PER_PAGE;
    }

    pSessionInfo->mappedUserBaseAddress = mappedUserBaseAddress;
    return status;
}

void UvmInitSessionInfo(UvmSessionInfo *pSessionInfo)
{
    pSessionInfo->pTargetCounterInfo = NULL;
    pSessionInfo->pidSessionOwner = UVM_PID_INIT_VALUE;
    pSessionInfo->euidTarget = UVM_ROOT_UID;
}

//
// Locking: this function uses sessionInfoLock
// 
UVM_STATUS NvUvmRemoveSessionInfo
(
    int sessionIndex,
    UvmProcessRecord *pCurrentProcessRecord
)
{
    UvmSessionInfo *pSessionInfo;
    UVM_STATUS status;

    down_write(&pCurrentProcessRecord->sessionInfoLock);

    status = NvUvmGetSessionInfo(sessionIndex,
                                 pCurrentProcessRecord,
                                 &pSessionInfo);
    if ( status == UVM_OK)
    {
        //
        // Free counter information if this is the only process using it.
        // This would happen if the target process has exited.
        //
        uvmlite_put_and_unrefcount_process_counter_info(
            pSessionInfo->pTargetCounterInfo);

        UvmInitSessionInfo(pSessionInfo);
    }
    up_write(&pCurrentProcessRecord->sessionInfoLock);

    return status;
}


UVM_STATUS NvUvmGetSessionCounterInfoSecure
(
    int sessionIndex,
    UvmProcessRecord *pCurrentProcessRecord,
    UvmProcessCounterInfo* *pProcessCounterInfo
)
{
    UVM_STATUS status;
    UvmSessionInfo *pSessionInfo;

    down_read(&pCurrentProcessRecord->sessionInfoLock);

    status = NvUvmGetSessionInfo(sessionIndex, 
                                 pCurrentProcessRecord,
                                 &pSessionInfo);
    if ( status == UVM_OK)
    {
        if (!UvmSecurityCheck(pSessionInfo->euidTarget))
        {
            up_read(&pCurrentProcessRecord->sessionInfoLock);
            return UVM_ERR_NOT_PERMITTED;
        }
        
        *pProcessCounterInfo = pSessionInfo->pTargetCounterInfo;
    }

    up_read(&pCurrentProcessRecord->sessionInfoLock);

    return status;
}

//
// Locking: this routine uses gpuHistoryLock
//
// Whoever gives a unique gpuUuid, gets unique index.
// Otherwise, he gets an index of existing entry, which matches gpuUuid
//
UVM_STATUS UvmGetGpuIndex
(
    UvmProcessCounterInfo *pProcessCounterInfo,
    UvmGpuUuid *gpuUuid, 
    int *gpuIndex
)
{
    unsigned index;

    WARN_ON(pProcessCounterInfo->numGpus > UVM_MAX_GPUS);

    down_write(&pProcessCounterInfo->gpuHistoryLock);
    for (index = 0; index < pProcessCounterInfo->numGpus; ++index)
    {
        UvmGpuUuid * pCurrentGpuUuid = &pProcessCounterInfo->gpuHistory[index];

        if (memcmp(pCurrentGpuUuid->uuid, gpuUuid->uuid,
                sizeof(pCurrentGpuUuid->uuid)) == 0)
        {
            // Found a matching gpuUuid
            break;
        }
    }
    //check if we can extend gpuHistoryList
    if (index == pProcessCounterInfo->numGpus 
    && pProcessCounterInfo->numGpus < UVM_MAX_GPUS)
    {
        memcpy(pProcessCounterInfo->gpuHistory[index].uuid, gpuUuid->uuid,
                sizeof(pProcessCounterInfo->gpuHistory[index].uuid));
        ++pProcessCounterInfo->numGpus;
    }
    
    up_write(&pProcessCounterInfo->gpuHistoryLock);
    
    if (index < UVM_MAX_GPUS)
    {
        *gpuIndex = index;
        return UVM_OK;
    }

    return UVM_ERR_INSUFFICIENT_RESOURCES;
}

//
// Allowing access only for the same user
// or root
//
UvmBool UvmSecurityCheck(uid_t euidTarget)
{
    return NV_CURRENT_EUID() == euidTarget || UVM_ROOT_UID == euidTarget;
}

unsigned UvmGetProcessId(void)
{
    return (unsigned) current->tgid;
}

UVM_STATUS UvmMapPage(struct page *page, void * mappingBaseAddress)
{
    unsigned long addr = (unsigned long) mappingBaseAddress;
    struct vm_area_struct *vma;
    UVM_STATUS status = UVM_ERR_INVALID_ARGUMENT;

    down_write(&current->mm->mmap_sem);
    vma = find_vma(current->mm, addr);
    if (vma)
    {
        if (vm_insert_page(vma, addr, page) == 0)
            status = UVM_OK;
    }
    up_write(&current->mm->mmap_sem);

    return status;
}

//
// Locking: Need to acquire process lock before incrementing process counters
//
void UvmIncrementProcessCounters
(
    UvmCounterInfo *procSingleGpuCounter, 
    UvmProcessRecord* pProcessRecord,
    UvmCounterName counterName,
    unsigned incrementVal
)
{
    unsigned long long *counterArray;
    // The value of the counter name is used as its index in the counter array
    unsigned counterIndex = counterName;

    // Increment process all gpu counters if any session has enabled them.
    
    if (NV_ATOMIC_READ(pProcessRecord->pProcessCounterInfo->
        procAllGpuCounter.sessionCount[counterIndex]) != 0)
    {
        counterArray = 
               pProcessRecord->pProcessCounterInfo->procAllGpuCounter.sysAddr;
        counterArray[counterIndex] = counterArray[counterIndex] + incrementVal;
    }
    

    // Process Single Gpu counters are enabled by default:
      
    counterArray = procSingleGpuCounter->sysAddr;
    counterArray[counterIndex] = counterArray[counterIndex] + incrementVal;
    
        
}

//
// This function returns the offset at which the counter value is present
//
static UVM_STATUS NvUvmGetCounterOffset
(
    unsigned long long *pUserMappedCounterPage,
    UvmCounterName counterName,
    void ** addr
)
{
    UVM_STATUS status;
    unsigned counterIndex;

    // Check if the counter name is valid and get its index
    status = NvUvmGetCounterIndex(counterName, &counterIndex);

    if(pUserMappedCounterPage != NULL && status == UVM_OK)
        *addr = &pUserMappedCounterPage[counterIndex];
    return status;
}

//
// Picks user VA for a given counter and returns offset of that counter
// Locking: this function uses sessionInfoLock
//
UVM_STATUS NvUvmMapCounter
(
    int sessionIndex,
    UvmCounterScope scope,
    UvmCounterName counterName,
    UvmGpuUuid gpuUuid,
    UvmProcessRecord *pCurrentProcessRecord,
    void ** addr
)
{
    unsigned long long *pUserMappedCounterPage;
    int gpuIndex;
    UVM_STATUS status;
    UvmSessionInfo *pSessionInfo;
    UvmProcessCounterInfo *pProcessCounterInfo = NULL;

    down_read(&pCurrentProcessRecord->sessionInfoLock);
    status = NvUvmGetSessionInfo(sessionIndex, 
                                 pCurrentProcessRecord,
                                 &pSessionInfo);
    up_read(&pCurrentProcessRecord->sessionInfoLock);

    pProcessCounterInfo = pSessionInfo->pTargetCounterInfo;

    if(status != UVM_OK)
        return status;

    switch (scope)
    {
        case UvmCounterScopeProcessSingleGpu:
            if(UvmGetGpuIndex(pProcessCounterInfo, &gpuUuid, &gpuIndex) != 
            UVM_OK)
                return UVM_ERR_INVALID_ARGUMENT;
 
            pUserMappedCounterPage = pSessionInfo->mappedUserBaseAddress + 
                                     gpuIndex * UVM_MAX_COUNTERS_PER_PAGE +
                                     UVM_PER_PROCESS_PER_GPU_COUNTERS_SHIFT;
            break;

        case UvmCounterScopeProcessAllGpu:
            pUserMappedCounterPage = pSessionInfo->mappedUserBaseAddress;
            break;

        case UvmCounterScopeGlobalSingleGpu:
            return UVM_ERR_NOT_SUPPORTED;
        default:
            return UVM_ERR_INVALID_ARGUMENT;
    }

    return NvUvmGetCounterOffset(pUserMappedCounterPage, counterName, addr);
}

/////////////////////////////////////////migrated from nvidia_uvm_counters!!!!!!

//
// Enables counters specified by the processId and config structure.
//
UVM_STATUS NvUvmCounterStateUpdate
(
    int sessionIndex,
    const UvmCounterConfig *config,
    unsigned count,
    UvmProcessRecord *pCurrentProcessRecord
)
{
    const UvmCounterConfig *currConfig;
    unsigned counterNum;
    UVM_STATUS status;
    UvmProcessCounterInfo *pProcessCounterInfo;

    status = NvUvmGetSessionCounterInfoSecure(sessionIndex, 
                                              pCurrentProcessRecord,
                                              &pProcessCounterInfo);
    if(status != UVM_OK)
        return status;

    if (count > UVM_MAX_COUNTERS_PER_IOCTL_CALL)
        return UVM_ERR_INVALID_ARGUMENT;

    for (counterNum = 0; counterNum < count; counterNum++)
    {
        currConfig = &config[counterNum];

        switch (currConfig->scope)
        {
            case UvmCounterScopeProcessSingleGpu:
                // These are enabled by default and cannot be disabled
                break;

            case UvmCounterScopeProcessAllGpu:
                if(currConfig->state == 
                UVM_COUNTER_CONFIG_STATE_ENABLE_REQUESTED)
                    status = NvUvmIncrementSessionCount(
                        &pProcessCounterInfo->procAllGpuCounter,
                        currConfig->name);
                else
                    status = NvUvmDecrementSessionCount(
                        &pProcessCounterInfo->procAllGpuCounter,
                        currConfig->name);
                break;

            case UvmCounterScopeGlobalSingleGpu:
                return UVM_ERR_NOT_SUPPORTED;
            default:
                return UVM_ERR_INVALID_ARGUMENT;
        }
    }

    return status;
}

UVM_STATUS NvUvmIncrementSessionCount
(
    UvmCounterInfo *pCtrInfo,
    UvmCounterName counterName
)
{
    unsigned counterIndex;
    UVM_STATUS status;

    status = NvUvmGetCounterIndex(counterName, &counterIndex);
    if(status == UVM_OK)
        NV_ATOMIC_INC(pCtrInfo->sessionCount[counterIndex]);

    return status;
}

UVM_STATUS NvUvmDecrementSessionCount
(
    UvmCounterInfo *pCtrInfo,
    UvmCounterName counterName
)
{
    unsigned counterIndex;
    UVM_STATUS status;

    status = NvUvmGetCounterIndex(counterName, &counterIndex);
    if(status == UVM_OK)
        NV_ATOMIC_DEC(pCtrInfo->sessionCount[counterIndex]);

    return status;
}

//
// this function checks if the countername is valid and returns its index
//
UVM_STATUS NvUvmGetCounterIndex
(
    UvmCounterName counterName,
    unsigned *counterIndex
)
{
    if(counterName >= 0 && counterName < UVM_TOTAL_COUNTERS) 
    {
        *counterIndex = counterName;
        return UVM_OK;
    }
    return UVM_ERR_INVALID_ARGUMENT;
}
